package de.hdm_stuttgart.mi.exam.unitmarking.reflect;

import org.junit.Assert;

import java.lang.reflect.*;
import java.util.Arrays;
import java.util.stream.Collectors;

public class ObjectWrapper<T> {

    private final Class<?> classz;
    private final T instance;

    static public void assertFinal(final Class<?> classZ, final String fieldName) {
        try {
            final int modifier = classZ.getDeclaredField(fieldName).getModifiers();
            Assert.assertTrue("»" + fieldName + "« can be modified, no immutable!", Modifier.isFinal(modifier));

        } catch (NoSuchFieldException e) {
            Assert.fail("No such field »" + fieldName + "«");
        }
    }

    public <R> void assertFieldExists(final Class<R> valueType, final String name) {
        try {
            final Field field = classz.getField(name);
            if (!valueType.equals(field.getType())) {
                Assert.fail("Field »" + name + "« in class »" + classz.getName() +  "« is not of type »" +
                        valueType.getName() + "«");
            }
        } catch (final NoSuchFieldException e) {
            Assert.fail("No such field »" + name + "« in class »" + classz.getName() + "«");
        }
    }

    public <R> R get(final Class<R> valueType, final String name) {

        try {
            final Field field = classz.getField(name);
            if (valueType.equals(field.getType())) {
                return (R) field.get(instance);
            } else {
                Assert.fail("Field »" + name + "« in class »" + classz.getName() +  "« is not of type »" +
                        valueType.getName() + "«");
            }
        } catch (final NoSuchFieldException e) {
            Assert.fail("No such field »" + name + "« in class »" + classz.getName() + "«");
        } catch (final IllegalAccessException e) {
            Assert.fail("Unable to access field »" + name + "« in class »" + classz.getName() + "«: " + e.getMessage());
        }
        return null;
    }
    public void set(final String name, final Object value) {

        try {
            final Field field = classz.getField(name);
            final Class<?>  argumentType = value.getClass();
            if (field.getType().equals(argumentType) ||
                    field.getType().equals(getPrimitiveType(argumentType))){
                field.set(instance, value);
            } else {
                Assert.fail("Field »" + name + "« in class »" + classz.getName() +  "« is not of type »" +
                        argumentType.getName() + "«");
            }
        } catch (final NoSuchFieldException e) {
            Assert.fail("No such field »" + name + "« in class »" + classz.getName() + "«");
        } catch (final IllegalAccessException e) {
            Assert.fail("Unable to access field »" + name + "« in class »" + classz.getName() + "«: " + e.getMessage());
        }
    }

    private Constructor<T> getConstructor(final Object ... parameter) {
        final Constructor<?>[] candidates = Arrays.
                stream(classz.getDeclaredConstructors()).
                filter(c -> matchesArgumentList(c, parameter)).
                toArray(Constructor<?>[]::new);
        if (0 == candidates.length) {
            Assert.fail("No suitable constructor »"  +
                    getSignature(classz.getSimpleName(), parameter) + "« found");
        } else if (1 < candidates.length) {
            Assert.fail("Multiple constructor matches due to ambiguous signature: " + Arrays.toString(candidates));
        }
        return (Constructor<T>) candidates[0];
    }

    public ObjectWrapper(final Class<T> classz, final Object ... parameter) {
        this.classz = classz;
        T tmpInstance = null;
        final Constructor<T> constructor = getConstructor(parameter);
        try {
            tmpInstance = constructor.newInstance(
                    Arrays.stream(parameter).toArray());
        } catch (final IllegalAccessException e) {
            Assert.fail("Unable to instantiate instance of class »" + classz.getName() + "«:\n" +
                    "Cannot access constructor " + constructor.toString());
        } catch (final InstantiationException|IllegalArgumentException|InvocationTargetException e) {
            Assert.fail("Unable to instantiate instance of class »" + classz.getName() + "«:\n  "+ e);
        }
        instance = tmpInstance;
    }

    public ObjectWrapper(final Class<T> classz, final Class<? extends Throwable> expectedException, final Object ... parameter) {
        this.classz = classz;
        T tmpInstance = null;
        final Constructor<T> constructor = getConstructor(parameter);
        try {
            tmpInstance = (T) constructor.newInstance(
                    Arrays.stream(parameter).toArray());
            Assert.fail("Expected exception of type »" + expectedException.getName() + "« to be thrown");
        } catch (final InstantiationException|IllegalAccessException|IllegalArgumentException|InvocationTargetException e) {
            if (e.getCause().getClass() != expectedException) {
                Assert.fail("Unable to instantiate: " + e + ", cause:\n  Expected exception of type »" +
                        expectedException.getName() + "« but was »"+ e.getCause().getClass().getName() + "«");
            }
        }
        instance = tmpInstance;
    }

    private Method getMethod(final Class<?> returnType, final String name, final Object ... parameter) {

        final Method[] candidates = Arrays.
                stream(classz.getDeclaredMethods()).
                filter(m-> m.getName().equals(name) && matchesArgumentList(m, parameter)).
                toArray(Method[]::new);

        if (0 == candidates.length) {
            Assert.fail("No suitable method »" + getSignature(returnType, name, parameter) +
                    "« found");
        } else if (1 < candidates.length) {
            Assert.fail("Multiple method matches due to ambiguous signature");
        }
        return candidates[0];
    }

    public <R> R invoke(final Class<R> returnType, final String name, final Object ... parameter) {

        final Method method = getMethod(returnType, name, parameter);
        if (method.getReturnType().equals(returnType)) {
            try {
                return (R) method.invoke(instance, parameter);
            } catch (final IllegalAccessException e) {
                Assert.fail("Unable to access method »" + method.toString() + "«: \n\t" + e.getMessage());
            }catch (final IllegalArgumentException|InvocationTargetException e) {
                Assert.fail("Unable to execute method: »" + method.toString() + "«\n" +
                        e + ", cause:" + e.getCause());
            }
        } else {
            Assert.fail("Method »" + method.getName() + "« does have return type »"  + method.getReturnType() + "«" +
                    ", expected »" + getSimplifiedTypeName(returnType) + "«");
        }
        return null;
    }

    public <R> R invoke(final Class<R> returnType, final String name,
                        final Class<? extends Throwable> expectedException, final Object ... parameter) {

        final Method method = getMethod(returnType, name, parameter);
        if (method.getReturnType().equals(returnType)) {
            try {
                R ret = (R) method.invoke(instance, parameter);
                Assert.fail("Expected exception of type »" + expectedException.getName() + "«");
                return ret;
            } catch (final IllegalAccessException| IllegalArgumentException|InvocationTargetException e) {
                if (e.getCause().getClass() != expectedException) {
                    Assert.fail("Unable to execute method: " + e + ", cause:\n  Expected exception of type »" +
                            expectedException.getName() + "« but was »"+ e.getCause().getClass().getName() + "«");
                }
            }
        } else {
            Assert.fail("Method »" + method.getName() + "« does have return type »"  + method.getReturnType() + "«" +
                    "rather then »" + returnType.getName() + "«");
        }
        return null;
    }

    // Constructor use
    private StringBuilder getSignature(final String name, final Object ... parameter) {

        final StringBuilder ret = new StringBuilder(classz.getName()).append("::");

        return ret.append(name).append('(')
                .append(Arrays.stream(parameter).map(this::appendObjectsType).collect(Collectors.joining(", ")))
                .append(')');
    }

    // Method use
    private <R> StringBuilder getSignature(final Class<R> returnType, final String name, final Object ... parameter) {

        return getSignature(name, parameter).insert(0, ' ').insert(0, returnType.getName());
    }

    private StringBuilder appendObjectsType (final Object object) {
        final StringBuilder builder = new StringBuilder();
        if (null == object) {
            builder.append("null");
        } else {
            appendType(object.getClass(), builder);
        }
        return builder;
    }

    static private String getSimplifiedTypeName(final Class<?> classz) {
        if (Byte.TYPE == classz) {
            return "byte";
        } else if (Short.TYPE == classz) {
            return "short";
        } else if (Integer.TYPE == classz) {
            return "int";
        } else if (Long.TYPE == classz) {
            return "long";
        } else if (Float.TYPE == classz) {
            return "float";
        } else if (Double.TYPE == classz) {
            return "double";
        } else if (Character.TYPE == classz) {
            return "char";
        } else if (Boolean.TYPE == classz) {
            return "boolean";
        } else {
            return classz.getName();
        }
    }

    private <R> void appendType(final Class<R> type, StringBuilder result) {

        if (type.isArray()) {
            final Class <?> componentType = type.getComponentType();
            result.append(getSimplifiedTypeName(componentType)).append("[]");
        } else {
            result.append(getSimplifiedTypeName(type));
        }
    }

    /**
     * Check for a given array of objects matching an {@link Executable}'s argument list.
     *
     * @param executable Method or constructor
     * @param parameter List
     * @return <code>true</code> if parameters match the {@link Executable}'s argument list, <code>false</code>
     * otherwise.
     */
    private boolean matchesArgumentList(final Executable executable, final Object ... parameter) {

        if (executable.getParameterCount() != parameter.length) {
            return false;
        } else {
            final Class<?>[] formalArgumentTypes = executable.getParameterTypes();
            for (int i = 0; i < formalArgumentTypes.length; i++) {
                final Class<?> parametersClass = parameter[i].getClass();

                if (!formalArgumentTypes[i].equals(parametersClass) &&
                        !formalArgumentTypes[i].equals(getPrimitiveType(parametersClass))) {
                    return false;
                }
            }
        }
        return true;
    }

    static private Class<?> getPrimitiveType (final Class<?> in) {
        if (in.equals(Byte.class)) {
            return byte.class;
        } else if (in.equals(Short.class)) {
            return short.class;
        } else if (in.equals(Integer.class)) {
            return int.class;
        } else if (in.equals(Long.class)) {
            return long.class;
        } else if (in.equals(Float.class)) {
            return float.class;
        } else if (in.equals(Double.class)) {
            return double.class;
        } else if (in.equals(Boolean.class)) {
            return boolean.class;
        } else if (in.equals(Character.class)) {
            return char.class;
        } else {
            return in;            // Type is no primitive
        }
    }

}
